
local MAJOR = "StarLibPluginUnitStats-1.0" 
local MINOR = 13
assert(LibStub, MAJOR.." requires LibStub") 
local LibUnitStats = LibStub:NewLibrary(MAJOR, MINOR)
if not LibUnitStats then return end
local self = LibUnitStats

if not LibUnitStats.__index then 
	LibUnitStats.__index = LibUnitStats
end

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")
local tooltip = CreateFrame("GameTooltip", "StarLibUnitStats", UIParent, "GameTooltipTemplate")
tooltip:SetOwner(UIParent, "ANCHOR_NONE")
local initialized
local frame = CreateFrame("Frame")

local factionList = {}

if not LibUnitStats.__index then
	LibUnitStats.__index = LibUnitStats
end

LibUnitStats.leftLines = {}
LibUnitStats.rightLines = {}

function initialize()
	if initialized then return end
	for i = 1, 20 do
		tooltip:AddDoubleLine(" ", " ")
		LibUnitStats.leftLines[i] = _G["StarLibUnitStatsTextLeft" .. i]
		LibUnitStats.rightLines[i] = _G["StarLibUnitStatsTextRight" .. i]
	end
	initialized = true
end

local function onEvent(frame, event)
	if event == "UPDATE_FACTION" then
		for i = 1, GetNumFactions() do
			local name = GetFactionInfo(i)
			factionList[name] = true
		end
	end
end

-- Create a new PluginUnitStats object
-- @name PluginUnitStats.New
-- @usage PluginUnitStats:New()
-- @return A new PluginUnitStats object
function LibUnitStats:New(environment)
	local obj = setmetatable({}, self)
	environment.GetUnitStats = self.GetUnitStats
	initialize()
	frame:SetScript("OnEvent", onEvent)
	frame:RegisterEvent("UPDATE_FACTION")
	return obj
end

-- Delete a PluginUnitStats object
-- @name PluginUnitStats.Del
-- @usage PluginUnitStats.Del()
-- @return Nothing
function LibUnitStats:Del()
	frame:SetScript("OnEvent", nil)
end

local getLocation, getGuild, getName

-- Return the default unit tooltip's stats
-- @usage LibUnitStats:GetValues()
-- @return Name, guild, and location
function LibUnitStats.GetUnitStats(unit)
	if not unit then unit = "mouseover" end
    if not UnitIsVisible(unit) or not UnitIsConnected(unit) then
        return nil
    end
	tooltip:ClearLines()
	tooltip:SetUnit(unit)
	return getName(), getGuild(), getLocation()
end

local LEVEL_start = "^" .. (type(LEVEL) == "string" and LEVEL or "Level")
function getLocation()
    local left_2 = self.leftLines[2]:GetText()
    local left_3 = self.leftLines[3]:GetText()
    if not left_2 or not left_3 then
        return nil
    end
    local hasGuild = not left_2:find(LEVEL_start)
    local factionText = not hasGuild and left_3 or self.leftLines[4]:GetText()
    if factionText == PVP then
        factionText = nil
    end
    local hasFaction = factionText and not UnitPlayerControlled("mouseover") and not UnitIsPlayer("mouseover") and (UnitFactionGroup("mouseover") or factionList[factionText])
    if hasGuild and hasFaction then
        return self.leftLines[4]:GetText()
    elseif hasGuild or hasFaction then
        return self.leftLines[5]:GetText()
    else
        return left_3
    end
end

function getGuild()
    local left_2 = self.leftLines[2]:GetText()
	if not left_2 then return "Err" end
    if left_2:find(LEVEL_start) then return nil end
    return "<" .. left_2 .. ">"
end

function getName()
	return self.leftLines[1]:GetText()
end